// DCF77 LED digital clock
// Author: Levente Maier 2012
//
//
// 										CPU 									Display digit
//                                   ATtiny2313
//                             ---------------------  								1111111111
//                            |                     |								8		16
//            Reset <---PA2---|1(/RESET/dW)   VCC 20|								8		16
//    LED Signal OK	<---PD0---|2(RXD)   (UCSK/SCL)19|---PB7--> 		 ISP			8		16
//    LED Seconds	<---PD1---|3(TXD)    (MISO/DO)18|---PB6--> LED g ISP 			8		16
//        			<---PA1---|4(XTAL2) (./DI/SDA)17|---PB5--> LED f ISP 			2222222222
//        			<---PA0---|5(XTAL1)     (OC1B)16|---PB4--> LED e				32		64
//        	 LED 1	<---PD2---|6(CKOUT/XCK) (OC1A)15|---PB3--> LED d				32		64
//           LED 2	<---PD3---|7(INT1)      (OC0A)14|---PB2--> LED c				32		64
//           LED 3	<---PD4---|8(T0)        (AIN1)13|---PB1--> LED b				32		64
//           LED 4	<---PD5---|9(OC0B/T1)   (AIN0)12|---PB0--> LED a				4444444444
//                            |10 GND        (ICP)11|---PD6--> DCF77 signal
//                             ---------------------

#define DCF77			6             // Port PD6 - DCF77

#define LED1            2 			  // Port PD2 - LED 10 	hours
#define LED2         	3             // Port PD3 - LED 	hours
#define LED3         	4             // Port PD4 - LED 10 	minutes
#define LED4	   		5             // Port PD5 - LED 	minutes

#define LEDa          	0             // Port PB0 - LEDa	top	
#define LEDb            1			  // Port PB1 - LEDb   	middle
#define LEDc			2             // Port PB2 - LEDc	bottom
#define LEDd			3             // Port PB3 - LEDd	left top
#define LEDe	      	4             // Port PB4 - LEDe	left bottom
#define LEDf			5             // Port PB5 - LEDf	right top
#define LEDg   			6             // Port PB6 - LEDg	right bottom

#define LEDSignalOk		0             // Port PD0 - LED	top
#define LEDSeconds   	1             // Port PD1 - LED	bottom


// In the test mode the counters for low or high signal will be displayed
#define testmode


#include <avr/io.h>
#include <avr/interrupt.h>

//Segment bit constant array 0->9
const uint8_t displaySegment[12]={1+4+8+16+32+64,16+64,1+16+2+32+4,1+2+4+16+64,2+8+16+64,1+2+4+8+64,1+2+4+8+32+64,1+16+64,1+2+4+8+16+32+64,1+2+4+8+16+64,0,2};

//during synchronization let the first and the second display digits rotate
#ifndef testmode
	const uint8_t coolDisplay1[8]={0	,0	,0	,	0	,4	,32	,8	,1};
	const uint8_t coolDisplay2[8]={1	,16	,64	,	4	,0	,0	,0	,0};
#endif

//Time
uint8_t min1 	= 0;	
uint8_t min10 	= 0;
uint8_t hour1 	= 0;	
uint8_t hour10 	= 0;

//time counter for no DCF77 signal
uint16_t minCount	= 0;

//rotate the digits
uint16_t coolDisplayCount = 0;

//DCF77 binary data array and his index
uint8_t recivedData[59];
uint8_t timeBit=0;

//Display
uint8_t displayDigitCounter=0;
uint8_t ledSecondsCounter=0;

//DCF77 Decoder
uint8_t signalOk=0;
uint8_t timeOk=0;
uint8_t lastSignal=0;
int16_t counterHigh;
int16_t counterLow;

#ifdef testmode
	uint8_t testvalue[4]={11,11,11,11};
#endif

/*
 * initialize mpu
 */
void Init(void)
{
	DDRD  &=  ~((1<<DCF77) );		//set DCF77 pin as input
	PORTD &=  ~((1<<DCF77) );		//deactivate the intern Pull-UP resistor for DCF77
	
	//set IO Ports D and B as output
    DDRD	|= ((1<<LED1) | (1<<LED2) | (1<<LED3) | (1<<LED4) | (1<<LEDSignalOk) | (1<<LEDSeconds));
	DDRB	|= ((1<<LEDa) | (1<<LEDb) | (1<<LEDc) | (1<<LEDd) | (1<<LEDe) | (1<<LEDf) | (1<<LEDg));

	//set all outputs to low
	PORTD = 0;
	PORTB = 127;
	
	//set interrupt controller and Timer1
	TIMSK  |= (1<<TOIE1);                                   // TOIE1=1 Interrupt enable     
    TCCR1B |= (0<<CS12) | (1<<CS11) | (1<<CS10);			// CS12=0, CS11=1, CS10=1 => Divide 64; 8000000Hz:64=125000 Hz       

    //enable all interrupts
	sei();
}

/*
 * display the time
 */
void DisplayTime(void)
{
	uint8_t aktualDigit=0;

	//switch all display digits off
	PORTD &= ~((1<<LED1) | (1<<LED2) | (1<<LED3) | (1<<LED4) );

	//first digit for 10 hours
	if (displayDigitCounter==0){
		if (timeOk==1)
			if (hour10<1)
				aktualDigit=0;		//display off don't display first 10hours digit of hour less then 10
			else
				aktualDigit=displaySegment[hour10];
		else
		#ifdef testmode
			aktualDigit=displaySegment[testvalue[3]];
		#else
			aktualDigit=coolDisplay1[coolDisplayCount];		//cool display if no valid time received
		#endif

		PORTD |=  (1<<LED1);
	}
	//second digit for hours
	if (displayDigitCounter==1) {
		if (timeOk==1)
			aktualDigit=displaySegment[hour1];
		else
		#ifdef testmode
			aktualDigit=displaySegment[testvalue[2]];
		#else
			aktualDigit=coolDisplay2[coolDisplayCount];		//cool display if no valid time received
		#endif
		PORTD |=  (1<<LED2);
	}
	//third digit for 10 minutes
	if (displayDigitCounter==2) {
		if (timeOk==1)
			aktualDigit=displaySegment[min10];
		else
			#ifdef testmode
				aktualDigit=displaySegment[testvalue[1]];
			#else
				aktualDigit=displaySegment[(timeBit / 10) % 10];	//display decoding time bit if no valid time received
				//aktualDigit=coolDisplay1[7-coolDisplayCount];
			#endif
		PORTD |=  (1<<LED3);
	}
	//fourth digit for minutes
	if (displayDigitCounter==3) {
		if (timeOk==1)
			aktualDigit=displaySegment[min1];
		else
			#ifdef testmode
				aktualDigit=displaySegment[testvalue[0]];
			#else
				aktualDigit=displaySegment[timeBit % 10]; //display decoding time bit if no valid time received
				//aktualDigit=coolDisplay2[7-coolDisplayCount];
			#endif
		PORTD |=  (1<<LED4);
	}

	//Display the actual digit
	PORTB=~(128+aktualDigit);

	//Let the second LEDs flashing
	if (ledSecondsCounter>62) {
		//both LEDs off
		PORTD &= ~(1<<LEDSeconds);
		PORTD &= ~(1<<LEDSignalOk);
	}
	else {
		//top LED on
		PORTD |=  (1<<LEDSeconds);
		//turn the LED for Signal OK on is the DCF77 signal OK both LEDs flashing
		if (signalOk==1) 	PORTD |=  (1<<LEDSignalOk);
	}

	//counter flash LED for seconds
	if ( ledSecondsCounter>125) {
		ledSecondsCounter=0;
	}

	//cool display if no valid time received
	if (((ledSecondsCounter % 5)==0) && (displayDigitCounter==0))
	{
		coolDisplayCount++;
		if (coolDisplayCount>7)
			coolDisplayCount=0;
	}

	//counter for display digits
	displayDigitCounter++;
	if (displayDigitCounter>=4) {
		displayDigitCounter=0;
		ledSecondsCounter++;
	}
}

/*
 * CheckParity
 */
uint8_t CheckParity(uint8_t arrayStart, uint8_t arrayEnd) {
	uint8_t i;
	uint8_t value=0;
	for (i=arrayStart;i<=arrayEnd;i++) value +=recivedData[i];
	if ( recivedData[arrayEnd+1] != (1 & value) )
		return 0;
	else
		return 1;

}

/*
 * Decode an verify the received array of bytes
 */
uint8_t Dcf77DecodeTime()
{
	uint8_t tmin1;
	uint8_t tmin10;
	uint8_t thour1;
	uint8_t thour10;

	//this Bit is always 1
	if ( recivedData[20] != 1 )
		return 0;

	//ParityCheck minute
	if (CheckParity(21,27)==0)
		return 0;

	//ParityCheck hour
	if (CheckParity(29,34)==0)
		return 0;

	//ParityCheck date
	//No important while not displayed
	//if (CheckParity(36,57)==0)
	//	return 0;

	tmin1=	recivedData[21]+(recivedData[22]<<1)+(recivedData[23]<<2)+(recivedData[24]<<3);
	tmin10=	recivedData[25]+(recivedData[26]<<1)+(recivedData[27]<<2);

	thour1=	recivedData[29]+(recivedData[30]<<1)+(recivedData[31]<<2)+(recivedData[32]<<3);
	thour10=recivedData[33]+(recivedData[34]<<1);

	//plausibility check
	if ( (tmin1<=9) && (tmin10<=5) && (thour10<=2) && (thour1<=9) && (thour1+10*thour10<=23))
	{
		min1=tmin1;
		min10=tmin10;
		hour1=thour1;
		hour10=thour10;
		timeOk=1;
		return 1;
	}

	return 0;

}

/*
 * Reset the time data array if no signal or pure signal detected
 */
void ResetTimeData()
{
	uint8_t i;
	for (i=0;i<59;i++) recivedData[i]=0;
	signalOk=0;
}

/*
 * increment the time with one minute
 */
void IncrementTime(void)
{
	min1++;
	if (min1==10) 	{ min1	=0; min10++; }
	if (min10==6) 	{ min10	=0; hour1++; }
	if (hour1==10) 	{ hour1	=0; hour10++;}
	if ((hour10==2) && (hour1==4)) { hour10=0; hour1=0; }
}

#ifdef testmode
	void setTestModeValue(uint16_t i) {
		if (i<=9999) {
			testvalue[3]=i / 1000;
			testvalue[2]=(i - testvalue[3]*1000) / 100;
			testvalue[1]=(i - testvalue[3]*1000 - testvalue[2]*100) / 10;
			testvalue[0]=i - testvalue[3]*1000 - testvalue[2]*100 - testvalue[1]*10;
		}
		else {
			testvalue[3]=10;
			testvalue[2]=10;
			testvalue[1]=10;
			testvalue[0]=10;
		}
	}
#endif

/*
 * Signal analysis
 */
void Dcf77decoder625Hz(void)
{
	//read DCF77 signal input port
	//and count the length of each signal level (one unit is 1/625 seconds)
	//The value 62-63 corresponds to 100ms a low DCF77 signal and
	//the value 125 corresponds to 200ms high DCF77 signal
	uint8_t signal;
	if (PIND & (1<<DCF77)) {
		signal=1; 
		counterHigh++; 
	}
	else { 
		signal=0;
		counterLow++;
	}


	//DCF77 signal changed from low to high (rising edge)
	if ((lastSignal==0) && (signal==1)  )
	{
		//the high signal is longer then 50ms
		if (counterHigh>31)
		{
			//detect data bit=0
			if ((counterHigh>31) && (counterHigh<70)  && (counterLow>450) && counterLow<1300) {
				recivedData[timeBit]=0;
				signalOk=1;timeBit++;
				#ifdef testmode
				// Show the counter of low signal should be between 58 and 64 ex: 00:62
					setTestModeValue(counterHigh);
				#endif
			}
			//detect data bit=1
			if ((counterHigh>110) && (counterHigh<140)  && (counterLow>450) && counterLow<1300) {
				recivedData[timeBit]=1;
				signalOk=1;timeBit++;
				#ifdef testmode
					// Show the counter of high signal should be between 120 and 130 ex: 01:23
					setTestModeValue(counterHigh);
				#endif
			}
			//detect end of time sequence counter low longer then one and half second
			if ((counterHigh<140) && (counterLow>1000) && counterLow<1300) {
				if (timeBit==59) {
					if (Dcf77DecodeTime()==1)
						minCount=0;				//received time is ok
					else
						IncrementTime();		//Parity or plausibility check failed

				}
				timeBit=0;
				signalOk=1;
			}
			//wrong signal while high signal > 300 ms, low signal longer then 2 seconds
			if ((counterHigh>200)  || (counterLow<450) || counterLow>1300) {
				#ifdef testmode
					// Error: Show the counter of low signal ex: 51:25
					setTestModeValue(counterHigh+ 5000);
				#endif
				ResetTimeData();
			}
			//reset signal counter
			counterLow=0; counterHigh=0;
			if (timeBit>=59 ) timeBit=0;

			//Synchronize the second flash light
			ledSecondsCounter=0;
		}
	}

	//No valid DCF77 radio signal and code since last minute detected
	if (++minCount==(uint16_t) 625*60) {
		IncrementTime();
		minCount=0;
	} 

	lastSignal=signal;	//save the actual signal level to detect rising edge
}


/*
 * Timer1 interrupt
 */
ISR(TIMER1_OVF_vect)
{
    unsigned char temp_sreg_Timer1;
    temp_sreg_Timer1 = SREG;			// save timer Status-Register

    TCNT1   = 65535UL-200UL;			// set Timer1 start value for Overflow-Interrupt
										// Interrupt frequency  = clock frequency/ prescaler / start value
										// 8.000.000 / 64 / 200 = 625 Hz

	Dcf77decoder625Hz();
	DisplayTime();

   	SREG = temp_sreg_Timer1;            // restore timer Status-Register
	return;
}


/*
 * Main program
 */
int main(void)
{
	Init();		//Initialize MPU
	while(1);	//infinite loop of do nothing
}
